# The point type

LineObjType := NewType(PictureObjectFamily, IsLineObj, [ "x1", "y1",
                                                         "x2", "y2" ]);

InstallGlobalFunction(LineObj,
    function(x1, y1, x2, y2)
        return Objectify(LineObjType, rec(x1 := x1,y1 := y1, x2 := x2,y2 := y2,
                                          marks := Set([])));
    end);

## some operations to make points easier to deal with
#InstallMethod(\+, "addition of two lines",
#    IsIdenticalObj, [IsLineObj, IsLineObj], 0,
#    function (p1, p2)
#        return LineObj(p1!.x + p2!.x, p1!.y + p2!.y);
#    end);

InstallMethod(\=, "equality operator for lines",
    IsIdenticalObj, [IsLineObj, IsLineObj], 0,
    function (p1, p2)
        return p1!.x = p2!.x and p1!.y = p2!.y and p1!.length ;
    end);

InstallMethod(\<, "less than for lines",
    IsIdenticalObj, [IsLineObj, IsLineObj], 0,
    function (p1, p2)
        return p1!.x1 < p2!.x1 or
               p1!.x1 = p2!.x1 and (p1!.y1 < p2!.y1 or
                   p1!.y1 = p2!.y1 and (p1!.x2 < p2!.x2 or
                       p1!.x2 = p2!.x2 and p1!.y2 < p2!.y2));
    end);

InstallMethod(PrintObj, "print LineObj object",
    true, [ IsLineObj ], 0,
    function (point)
        Print("LineObj( ", point!.x1, ", ", point!.y1,
              ", " point!.x2, ", ", point!.y2, " )");
    end);

# PictureObject methods
InstallMethod(SquareDistance, "close objects implementation for line",
    true, [ IsLineObj, IsLineObj ], 0,
    function (y, p)
        return (p!.x - y!.x)^2 + (p!.y - y!.y)^2;
    end);

InstallMethod(DrawObject, "draw a point",
    true, [ IsLineObj, IsGraphicSheet ], 0,
    function (p, gs)
	return Line(gs, Int(p!.x1), Int(p!.y1),
                    Int(p!.x2 - p!.x1), Int(p!.y2 - p!.y1));
    end);


#### The LineConfiguration object
LineConfigurationType := NewType(ConfigurationFamily, IsLineConfiguration,
                                  [ "gwin", "owin", "lines" ]);

InstallGlobalFunction(LineConfiguration,
    function(arg)
        local x, y, xdim, ydim, ox, oy, oxdim, oydim, len;

        len := Length(arg);
        if len <> 4 and len <> 8 then
            Error("must pass 4 or 8 arguments");
        fi;
        x := arg[1];
        y := arg[2];
        xdim := arg[3];
        ydim := arg[4];
        if len = 8 then
            ox := arg[5];
            oy := arg[6];
            oxdim := arg[7];
            oydim := arg[8];
        else
            ox := x;
            oy := y;
            oxdim := xdim;
            oydim := ydim;
        fi;
        return Objectify(LineConfigurationType, rec(
            gwin := rec(x:=x, y:=y, xdim:=xdim, ydim:=ydim),
            owin := rec(x:=ox, y:=oy, xdim:=oxdim, ydim:=oydim),
            lines := Set([])
        ));
    end);

# other methods ...
InstallMethod(PrintObj, "the print function for point configurations",
    true, [ IsLineConfiguration ], 0,
    function (pcnf)
        Print("LineConfiguration( ",
              pcnf!.gwin.x, ", ", pcnf!.gwin.y, ", ",
              pcnf!.gwin.xdim, ", ", pcnf!.gwin.ydim, ", ",
              pcnf!.owin.x, ", ", pcnf!.owin.y, ", ",
              pcnf!.owin.xdim, ", ", pcnf!.owin.ydim, " )");
    end);

# Configuration methods
InstallMethod(ConfigWindowArea, "the area of the window",
    true, [ IsLineConfiguration ], 0,
    function (pcnf)
        return pcnf!.gwin.xdim * pcnf!.gwin.ydim;
    end);

InstallMethod(RandomNewObject, "create new object",
    true, [ IsLineConfiguration, IsFunction ], 0,
    function (pcnf, rnd)
        return LineObj(pcnf!.gwin.x + rnd() * pcnf!.gwin.xdim,
                     pcnf!.gwin.y + rnd() * pcnf!.gwin.ydim);
    end);
InstallMethod(ChooseRandomObject, "choose a random point",
    true, [ IsLineConfiguration ], 0,
    function (pcnf)
        return Random(pcnf!.lines);
    end);
InstallMethod(AddObject, "add object to point configuration",
    true, [ IsLineConfiguration, IsLineObj ], 0,
    function (pcnf, pt)
        AddSet(pcnf!.lines, pt);
    end);
InstallMethod(DelObject, "delete object to point configuration",
    true, [ IsLineConfiguration, IsLineObj ], 0,
    function (pcnf, pt)
        RemoveSet(pcnf!.lines, pt);
    end);
InstallMethod(CountObjects, "number of objects in configuration",
    true, [ IsLineConfiguration ], 0,
    function (pcnf)
        return Length(pcnf!.lines);
    end);
InstallMethod(CloseObjects, "close objects in configuration",
    true, [ IsLineConfiguration, IsLineObj, IsScalar ], 0,
    function (pcnf, pt, r)
        local r2, num, p;

        r2 := r^2;
        num := 0;
        for p in pcnf!.lines do
            if pt <> p and SquareDistance(pt, p) <= r2 then
                num := num + 1;
            fi;
        od;
        return num;
    end);

InstallMethod(CountObjectsWithMark, "get the list of objects with a mark",
    true, [ IsLineConfiguration, IsString ], 0,
    function (pcnf, mark)
        return Length(GetObjectsWithMark(pcnf, mark));
    end);
InstallMethod(GetObjectsWithMark, "get the list of objects with a mark",
    true, [ IsLineConfiguration, IsString ], 0,
    function (pcnf, mark)
        return Filtered(pcnf!.lines, pt -> HasMark(pt, mark));
    end);
InstallMethod(CloseObjectsWithMark, "close objects in cnf with a mark",
    true, [ IsLineConfiguration, IsLineObj, IsScalar, IsString ], 0,
    function (pcnf, pt, r, mark)
        local r2, num, p;

        r2 := r^2;
        num := 0;
        for p in pcnf!.lines do
            if pt <> p and SquareDistance(pt, p) <= r2 and HasMark(p,mark) then
                num := num + 1;
            fi;
        od;
        return num;
    end);

InstallMethod(DrawConfiguration, "draw configuration",
    true, [ IsLineConfiguration, IsGraphicSheet ], 0,
    function (pcnf, gs)
        local p;
        for p in pcnf!.lines do
            DrawObject(p, gs);
        od;
    end);


