# The point type

PointType := NewType(PictureObjectFamily, IsPoint, [ "x", "y" ]);

InstallGlobalFunction(Point,
    function(x, y)
        return Objectify(PointType, rec(x := x, y := y, marks := Set([])));
    end);

# some operations to make points easier to deal with
InstallMethod(\+, "addition of two points",
    IsIdenticalObj, [IsPoint, IsPoint], 0,
    function (p1, p2)
        return Point(p1!.x + p2!.x, p1!.y + p2!.y);
    end);

InstallMethod(\=, "equality operator for points",
    IsIdenticalObj, [IsPoint, IsPoint], 0,
    function (p1, p2)
        return p1!.x = p2!.x and p1!.y = p2!.y;
    end);

InstallMethod(\<, "less than for points",
    IsIdenticalObj, [IsPoint, IsPoint], 0,
    function (p1, p2)
        return p1!.x < p2!.x or (p1!.x = p2!.x and p1!.y < p2!.y);
    end);

InstallMethod(PrintObj, "print Point object",
    true, [ IsPoint ], 0,
    function (point)
        Print("Point( ", point!.x, ", ", point!.y, " )");
    end);

# PictureObject methods
InstallMethod(SquareDistance, "close objects implementation for points",
    true, [ IsPoint, IsPoint ], 0,
    function (y, p)
        return (p!.x - y!.x)^2 + (p!.y - y!.y)^2;
    end);

InstallMethod(DrawObject, "draw a point",
    true, [ IsPoint, IsGraphicSheet ], 0,
    function (p, gs)
        return Disc(gs, Int(p!.x), Int(p!.y), 3);
    end);


#### The PointConfiguration object
PointConfigurationType := NewType(ConfigurationFamily, IsPointConfiguration,
                                  [ "gwin", "owin", "points" ]);

InstallGlobalFunction(PointConfiguration,
    function(arg)
        local x, y, xdim, ydim, ox, oy, oxdim, oydim, len;

        len := Length(arg);
        if len <> 4 and len <> 8 then
            Error("must pass 4 or 8 arguments");
        fi;
        x := arg[1];
        y := arg[2];
        xdim := arg[3];
        ydim := arg[4];
        if len = 8 then
            ox := arg[5];
            oy := arg[6];
            oxdim := arg[7];
            oydim := arg[8];
        else
            ox := x;
            oy := y;
            oxdim := xdim;
            oydim := ydim;
        fi;
        return Objectify(PointConfigurationType, rec(
            gwin := rec(x:=x, y:=y, xdim:=xdim, ydim:=ydim),
            owin := rec(x:=ox, y:=oy, xdim:=oxdim, ydim:=oydim),
            points := Set([])
        ));
    end);

# other methods ...
InstallMethod(PrintObj, "the print function for point configurations",
    true, [ IsPointConfiguration ], 0,
    function (pcnf)
        Print("PointConfiguration( ",
              pcnf!.gwin.x, ", ", pcnf!.gwin.y, ", ",
              pcnf!.gwin.xdim, ", ", pcnf!.gwin.ydim, ", ",
              pcnf!.owin.x, ", ", pcnf!.owin.y, ", ",
              pcnf!.owin.xdim, ", ", pcnf!.owin.ydim, " )");
    end);

# Configuration methods
InstallMethod(ConfigWindowArea, "the area of the window",
    true, [ IsPointConfiguration ], 0,
    function (pcnf)
        return pcnf!.gwin.xdim * pcnf!.gwin.ydim;
    end);

InstallMethod(RandomNewObject, "create new object",
    true, [ IsPointConfiguration, IsFunction ], 0,
    function (pcnf, rnd)
        return Point(pcnf!.gwin.x + rnd() * pcnf!.gwin.xdim,
                     pcnf!.gwin.y + rnd() * pcnf!.gwin.ydim);
    end);
InstallMethod(ChooseRandomObject, "choose a random point",
    true, [ IsPointConfiguration ], 0,
    function (pcnf)
        return Random(pcnf!.points);
    end);
InstallMethod(AddObject, "add object to point configuration",
    true, [ IsPointConfiguration, IsPoint ], 0,
    function (pcnf, pt)
        AddSet(pcnf!.points, pt);
    end);
InstallMethod(DelObject, "delete object to point configuration",
    true, [ IsPointConfiguration, IsPoint ], 0,
    function (pcnf, pt)
        RemoveSet(pcnf!.points, pt);
    end);
InstallMethod(CountObjects, "number of objects in configuration",
    true, [ IsPointConfiguration ], 0,
    function (pcnf)
        return Length(pcnf!.points);
    end);
InstallMethod(CloseObjects, "close objects in configuration",
    true, [ IsPointConfiguration, IsPoint, IsScalar ], 0,
    function (pcnf, pt, r)
        local r2, num, p;

        r2 := r^2;
        num := 0;
        for p in pcnf!.points do
            if pt <> p and SquareDistance(pt, p) <= r2 then
                num := num + 1;
            fi;
        od;
        return num;
    end);

InstallMethod(CountObjectsWithMark, "get the list of objects with a mark",
    true, [ IsPointConfiguration, IsString ], 0,
    function (pcnf, mark)
        return Length(GetObjectsWithMark(pcnf, mark));
    end);
InstallMethod(GetObjectsWithMark, "get the list of objects with a mark",
    true, [ IsPointConfiguration, IsString ], 0,
    function (pcnf, mark)
        return Filtered(pcnf!.points, pt -> HasMark(pt, mark));
    end);
InstallMethod(CloseObjectsWithMark, "close objects in cnf with a mark",
    true, [ IsPointConfiguration, IsPoint, IsScalar, IsString ], 0,
    function (pcnf, pt, r, mark)
        local r2, num, p;

        r2 := r^2;
        num := 0;
        for p in pcnf!.points do
            if pt <> p and SquareDistance(pt, p) <= r2 and HasMark(p,mark) then
                num := num + 1;
            fi;
        od;
        return num;
    end);

InstallMethod(DrawConfiguration, "draw configuration",
    true, [ IsPointConfiguration, IsGraphicSheet ], 0,
    function (pcnf, gs)
        local p;
        for p in pcnf!.points do
            ColourObject(p, DrawObject(p, gs));
        od;
    end);


